/*
 *  CDEmuD: Daemon object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "cdemud.h"
#include "cdemud-service-glue.h"
#include "cdemud-marshallers.h"


/******************************************************************************\
 *                                 Misc stuff                                 *
\******************************************************************************/
/* Nice macro that gives for loop on GList */
#define G_LIST_FOR_EACH(cursor,list) \
    for ((cursor) = (list); (cursor); (cursor) = (cursor)->next)
    
#define SANITY_CHECK()                                          \
    if (!self) {                                                \
        cdemud_error(CDEMUD_E_OBJISNULL, error);                \
        return FALSE;                                           \
    }                                                           \
    if (!CDEMUD_IS_DAEMON(self)) {                              \
        cdemud_error(CDEMUD_E_INVALIDOBJ, error);               \
        return FALSE;                                           \
    }                                                           \
    if (CDEMUD_DAEMON_GET_PRIVATE(self)->dispose_has_run) {     \
        cdemud_error(CDEMUD_E_OBJDISPOSED, error);              \
        return FALSE;                                           \
    }                                                           \
    if (!CDEMUD_DAEMON_GET_PRIVATE(self)->initialized) {        \
        cdemud_error(CDEMUD_E_OBJNOTINIT, error);               \
        return FALSE;                                           \
    }    

#define CHECK_ARG(param)                                            \
    if (!param) {                                                   \
        cdemud_error(CDEMUD_E_INVALIDARG, error);                   \
        return FALSE;                                               \
    }

    
/******************************************************************************\
 *                              Private structure                             *
\******************************************************************************/
#define CDEMUD_DAEMON_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), CDEMUD_TYPE_DAEMON, CDEMUD_DaemonPrivate))

typedef struct _CDEMUD_DaemonPrivate CDEMUD_DaemonPrivate;

struct _CDEMUD_DaemonPrivate {
    gboolean dispose_has_run;
    
    gboolean initialized;
    
    gchar *version;
    gchar *audio_backend;
    
    GObject *mirage;
    
    GMainLoop *main_loop;
    
    DBusGConnection *bus;
    
    struct sysfs_driver *sysfs_driver;
    
    gint number_of_devices;
    GList *list_of_devices;    
};


/******************************************************************************\
 *                              Private functions                             *
\******************************************************************************/
static void __cdemud_daemon_device_change_handler (GObject *device, gint change_type, CDEMUD_Daemon *self) {
    /*CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);*/
    gint number = 0;
        
    cdemud_device_get_device_number(CDEMUD_DEVICE(device), &number, NULL);
    g_signal_emit_by_name(self, "device-change", number, change_type, NULL);
    
    return;
}

static gboolean __cdemud_daemon_destroy_devices (CDEMUD_Daemon *self, GError **error) {
    SANITY_CHECK();
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    
    GList *entry = NULL;
    
    G_LIST_FOR_EACH(entry, _priv->list_of_devices) {
        GObject *dev = entry->data;
        if (dev) {
            g_object_unref(dev);
        }
    }
    
    g_list_free(_priv->list_of_devices);
    
    return TRUE;
}

static gboolean __cdemud_daemon_get_device (CDEMUD_Daemon *self, gint device_number, GObject **device, GError **error) {
    SANITY_CHECK();
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    
    CHECK_ARG(device);
    
    if (device_number >= 0 && device_number < _priv->number_of_devices) {
        *device = g_list_nth_data(_priv->list_of_devices, device_number);
    } else {
        cdemud_error(CDEMUD_E_INVALIDDEVICE, error);
        return FALSE;
    }
    
    return TRUE;
}

    
/******************************************************************************\
 *                                 Public API                                 *
\******************************************************************************/
gboolean cdemud_daemon_initialize (CDEMUD_Daemon *self, gchar *audio_backend, gchar *audio_device, GError **error) {
    CDEMUD_DaemonPrivate *_priv = NULL;
    
    /* Do the error checking */
    if (!self) {
        cdemud_error(CDEMUD_E_OBJISNULL, error);
        return FALSE;
    }
    if (!CDEMUD_IS_DAEMON(self)) {
        cdemud_error(CDEMUD_E_INVALIDOBJ, error);
        return FALSE;
    }
    if (CDEMUD_DAEMON_GET_PRIVATE(self)->dispose_has_run) {
        cdemud_error(CDEMUD_E_OBJDISPOSED, error);
        return FALSE;
    }
    
    /* Start initialization */
    _priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    
    DBusGProxy *bus_proxy = NULL;
    guint result = 0;
    struct dlist *list_devices = NULL;
    struct sysfs_device *sysfs_device = NULL;
    
    /* Mirage */
    _priv->mirage = g_object_new(MIRAGE_TYPE_MIRAGE, NULL);
        
    /* Glib's main loop */
    _priv->main_loop = g_main_loop_new(NULL, FALSE);
    
    /* Open driver */    
    _priv->sysfs_driver = sysfs_open_driver("cdemu-bus", "cdemu-driver");
    if (!_priv->sysfs_driver) {
        cdemud_error(CDEMUD_E_NODRIVER, error);
        return FALSE;
    }

    /* Get devices list; for each device in it, create device object */
    list_devices = sysfs_get_driver_devices(_priv->sysfs_driver);
    if (!list_devices) {
        cdemud_error(CDEMUD_E_NODEVICES, error);
        return FALSE;
    }
    
    dlist_for_each_data(list_devices, sysfs_device, struct sysfs_device) {
        GObject *dev = g_object_new(CDEMUD_TYPE_DEVICE, NULL);
        
        /* Initialize device */
        if (cdemud_device_initialize(CDEMUD_DEVICE(dev), _priv->number_of_devices, sysfs_device, audio_backend, audio_device, _priv->mirage, NULL)) {
            /* Set parent */
            mirage_object_set_parent(MIRAGE_OBJECT(dev), G_OBJECT(self), NULL);
            /* Don't attach child... MIRAGE_Objects pass debug context to children,
               and CDEMUD_Devices have each its own context... */
            /* Add media changed notification signal handling */
            g_signal_connect(dev, "device-change", (GCallback)__cdemud_daemon_device_change_handler, self);
            /* Add it to devices list */
            _priv->list_of_devices = g_list_append(_priv->list_of_devices, dev);
            _priv->number_of_devices++;
        } else {
            g_object_unref(dev);
            cdemud_error(CDEMUD_E_DEVICEINITFAILED, error);
            return FALSE;
        }            
    }

    /* Initialize our DBUS interface */        
    dbus_g_object_type_install_info(CDEMUD_TYPE_DAEMON, &dbus_glib_cdemud_daemon_object_info);
    _priv->bus = dbus_g_bus_get(DBUS_BUS_SYSTEM, error);
    if (!_priv->bus) {
        cdemud_error(CDEMUD_E_DBUSCONNECT, error);
        return FALSE;
    }
    
    bus_proxy = dbus_g_proxy_new_for_name(_priv->bus, "org.freedesktop.DBus", "/org/freedesktop/DBus", "org.freedesktop.DBus");
    
    if (!dbus_g_proxy_call(bus_proxy, "RequestName", error, G_TYPE_STRING, "net.sf.cdemu.CDEMUD_Daemon", G_TYPE_UINT, 0, G_TYPE_INVALID, G_TYPE_UINT, &result, G_TYPE_INVALID)) {
        cdemud_error(CDEMUD_E_DBUSNAMEREQUEST, error);
        return FALSE;
    }

    dbus_g_connection_register_g_object(_priv->bus, "/CDEMUD_Daemon", G_OBJECT(self));
    
    /* We successfully finished initialization */
    _priv->initialized = TRUE;
    
    return TRUE;
}

gboolean cdemud_daemon_start_daemon (CDEMUD_Daemon *self, GError **error) {
    SANITY_CHECK();
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);

    g_signal_emit_by_name(G_OBJECT(self), "daemon-change", CDEMUD_DAEMON_CHANGE_STARTED, NULL);
    g_main_loop_run(_priv->main_loop);
    
    return TRUE;
}

gboolean cdemud_daemon_stop_daemon (CDEMUD_Daemon *self, GError **error) {
    SANITY_CHECK();        
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);

    g_main_loop_quit(_priv->main_loop);
    g_signal_emit_by_name(G_OBJECT(self), "daemon-change", CDEMUD_DAEMON_CHANGE_STOPPED, NULL);
    
    return TRUE;
}


/******************************************************************************\
 *                           DBUS interface functions                         *
\******************************************************************************/
gboolean cdemud_daemon_hail (CDEMUD_Daemon *self, GError **error) {
    SANITY_CHECK();
    /*DEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);*/

    return TRUE;
}

static gboolean __cdemud_daemon_get_version_module (CDEMUD_Daemon *self, gchar **version, GError **error) {
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    struct sysfs_attribute *sys_version = NULL;
        
    /* Read version */
    sys_version = sysfs_get_driver_attr(_priv->sysfs_driver, "version");
    if (sys_version) {
        if (!sysfs_read_attribute(sys_version)) {
            *version = g_strdup(sys_version->value);
        } else {
            cdemud_error(CDEMUD_E_ATTRREAD, error);
            return FALSE;
        }
    } else {
        cdemud_error(CDEMUD_E_ATTROPEN, error);
        return FALSE;
    }
    
    return TRUE;
}

static gboolean __cdemud_daemon_get_version_library (CDEMUD_Daemon *self, gchar **version, GError **error) {
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    /* Read version */
    return mirage_mirage_get_version(MIRAGE_MIRAGE(_priv->mirage), version, error);
}

static gboolean __cdemud_daemon_get_version_daemon (CDEMUD_Daemon *self, gchar **version, GError **error) {
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    /* Copy version string */
    *version = g_strdup(_priv->version);
    return TRUE;
}

gboolean cdemud_daemon_get_version (CDEMUD_Daemon *self, gchar *type, gchar **version, GError **error) {
    SANITY_CHECK(); 
    
    CHECK_ARG(type);
    CHECK_ARG(version);
    
    if (!g_ascii_strcasecmp(type, "daemon")) {
        /* Return daemon's version */
        return __cdemud_daemon_get_version_daemon(self, version, error);
    } else if (!g_ascii_strcasecmp(type, "library")) {
        /* Return libary's version */
        return __cdemud_daemon_get_version_library(self, version, error);
    } else if (!g_ascii_strcasecmp(type, "module")) {
        /* Return module's version */
        return __cdemud_daemon_get_version_module(self, version, error);
    }
    
    /* Invalid type */
    cdemud_error(CDEMUD_E_INVALIDARG, error);
    return FALSE;    
}


static gboolean __cdemud_daemon_get_device_debug_masks_module (CDEMUD_Daemon *self, GPtrArray **masks, GError **error) {
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    struct sysfs_attribute *sys_debug_masks = NULL;
    gint i = 0;
        
    *masks = g_ptr_array_new();
    
    /* Read debug masks */
    sys_debug_masks = sysfs_get_driver_attr(_priv->sysfs_driver, "debug_masks");
    if (sys_debug_masks) {
        if (!sysfs_read_attribute(sys_debug_masks)) {
            /* Split whole output in lines */
            gchar **mask_lines = g_strsplit(sys_debug_masks->value, ";\n", 0);
            for (i = 0; mask_lines[i]; i++) {
                /* Check if we didn't get empty line (the remainder...) */
                if (strcmp(mask_lines[i], "")) {
                    /* Split line into mask name and mask value */
                    gchar **mask_str = g_strsplit(mask_lines[i], ":", 0);
                    
                    /* Create value array */
                    GValueArray *mask_entry = g_value_array_new(2);
                    /* Mask name */
                    g_value_array_append(mask_entry, NULL);
                    g_value_init(g_value_array_get_nth(mask_entry, 0), G_TYPE_STRING);
                    g_value_set_string(g_value_array_get_nth(mask_entry, 0), mask_str[0]);
                    /* Mask value */
                    g_value_array_append(mask_entry, NULL);
                    g_value_init(g_value_array_get_nth(mask_entry, 1), G_TYPE_INT);
                    g_value_set_int(g_value_array_get_nth(mask_entry, 1), g_strtod(mask_str[1], NULL));
                    /* Add mask's value array to masks' pointer array */
                    g_ptr_array_add(*masks, mask_entry);
                    
                    g_strfreev(mask_str);
                }
            }
            
            g_strfreev(mask_lines);
        } else {
            cdemud_error(CDEMUD_E_ATTRREAD, error);
            return FALSE;
        }
    } else {
        cdemud_error(CDEMUD_E_ATTROPEN, error);
        return FALSE;
    }
        
    return TRUE;
}

static gboolean __cdemud_daemon_get_device_debug_masks_daemon (CDEMUD_Daemon *self, GPtrArray **masks, GError **error) {
    /*CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);*/
    gint i = 0;
    
    *masks = g_ptr_array_new();
    
    struct {
        gint mask_value;
        gchar *mask_name;
    } dbg_masks[] = {
        { DAEMON_DEBUG_DEV_PC_FIXME, "DAEMON_DEBUG_DEV_PC_FIXME" },
        { DAEMON_DEBUG_DEV_PC_TRACE, "DAEMON_DEBUG_DEV_PC_TRACE" },
        { DAEMON_DEBUG_DEV_PC_DUMP, "DAEMON_DEBUG_DEV_PC_DUMP" },
        { DAEMON_DEBUG_DEV_AUDIOPLAY,"DAEMON_DEBUG_DEV_AUDIOPLAY" },
    };
    
    for (i = 0; i < G_N_ELEMENTS(dbg_masks); i++) {
        /* Create value array */
        GValueArray *mask_entry = g_value_array_new(2);
        /* Mask name */
        g_value_array_append(mask_entry, NULL);
        g_value_init(g_value_array_get_nth(mask_entry, 0), G_TYPE_STRING);
        g_value_set_string(g_value_array_get_nth(mask_entry, 0), dbg_masks[i].mask_name);
        /* Mask value */
        g_value_array_append(mask_entry, NULL);
        g_value_init(g_value_array_get_nth(mask_entry, 1), G_TYPE_INT);
        g_value_set_int(g_value_array_get_nth(mask_entry, 1), dbg_masks[i].mask_value);
        /* Add mask's value array to masks' pointer array */
        g_ptr_array_add(*masks, mask_entry);
    }
    
    return TRUE;
}

static gboolean __cdemud_daemon_get_device_debug_masks_library (CDEMUD_Daemon *self, GPtrArray **masks, GError **error) {
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    /* Relay to libMirage */
    return mirage_mirage_get_supported_debug_masks(MIRAGE_MIRAGE(_priv->mirage), masks, error);
}

gboolean cdemud_daemon_get_device_debug_masks (CDEMUD_Daemon *self, gchar *type, GPtrArray **masks, GError **error) {
    SANITY_CHECK();
    
    CHECK_ARG(type);
    CHECK_ARG(masks);
    
    if (!g_ascii_strcasecmp(type, "daemon")) {
        /* Return daemon's debug masks */
        return __cdemud_daemon_get_device_debug_masks_daemon(self, masks, error);
    } else if (!g_ascii_strcasecmp(type, "module")) {
        /* Return module's debug masks */
        return __cdemud_daemon_get_device_debug_masks_module(self, masks, error);
    } else if (!g_ascii_strcasecmp(type, "library")) {
        /* Return library's debug masks */
        return __cdemud_daemon_get_device_debug_masks_library(self, masks, error);
    }
    
    /* Invalid type */
    cdemud_error(CDEMUD_E_INVALIDARG, error);
    return FALSE;
}

static void __cdemud_daemon_add_supported_image (gpointer data, gpointer user_data) {
    MIRAGE_PluginInfo_Parser* plugin_info = data;
    GPtrArray* supported_images_array = user_data;
    
    GValueArray *image_entry = NULL;
    
    /* Create value array for supported image */
    image_entry = g_value_array_new(3);
    /* Parser ID */
    g_value_array_append(image_entry, NULL);
    g_value_init(g_value_array_get_nth(image_entry, 0), G_TYPE_STRING);
    g_value_set_string(g_value_array_get_nth(image_entry, 0), plugin_info->image_id);
    
    /* Suffixes */
    g_value_array_append(image_entry, NULL);
    g_value_init(g_value_array_get_nth(image_entry, 1), G_TYPE_STRV);
    g_value_set_boxed(g_value_array_get_nth(image_entry, 1), plugin_info->suffixes);
    
    /* Description */
    g_value_array_append(image_entry, NULL);
    g_value_init(g_value_array_get_nth(image_entry, 2), G_TYPE_STRING);
    g_value_set_string(g_value_array_get_nth(image_entry, 2), plugin_info->description);
    /* Add mask's value array to masks' pointer array */
    g_ptr_array_add(supported_images_array, image_entry);
    
    return;
}

gboolean cdemud_daemon_get_supported_image_types (CDEMUD_Daemon *self, GPtrArray **image_types, GError **error) {
    SANITY_CHECK();
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    
    CHECK_ARG(image_types);
    
    *image_types = g_ptr_array_new();
    
    mirage_mirage_for_each_plugin(MIRAGE_MIRAGE(_priv->mirage), MIRAGE_PLUGIN_PARSER, __cdemud_daemon_add_supported_image, *image_types, NULL);
    
    return TRUE;
}


gboolean cdemud_daemon_get_number_of_devices (CDEMUD_Daemon *self, gint *number_of_devices, GError **error) {
    SANITY_CHECK();
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    
    CHECK_ARG(number_of_devices);
    *number_of_devices = _priv->number_of_devices;
    
    return TRUE;
}

gboolean cdemud_daemon_device_get_status (CDEMUD_Daemon *self, gint device_number, gboolean *loaded, gchar **image_type, gchar ***file_names, GError **error) {
    SANITY_CHECK();
    /*CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);*/
    GObject *dev = NULL;

    if (__cdemud_daemon_get_device(self, device_number, &dev, error)) {
        if (cdemud_device_get_status(CDEMUD_DEVICE(dev), loaded, image_type, file_names, error)) {
            return TRUE;
        }
    }
    
    return FALSE;
}

gboolean cdemud_daemon_device_load (CDEMUD_Daemon *self, gint device_number, gchar **file_names, GError **error) {
    SANITY_CHECK();
    /*CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);*/
    GObject *dev = NULL;
        
    if (__cdemud_daemon_get_device(self, device_number, &dev, error)) {
        if (cdemud_device_load_disc(CDEMUD_DEVICE(dev), file_names, error)) {
            return TRUE;
        }
    }
    
    return FALSE;
}

gboolean cdemud_daemon_device_unload (CDEMUD_Daemon *self, gint device_number, GError **error) {
    SANITY_CHECK();
    /*CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);*/
    GObject *dev = NULL;
        
    if (__cdemud_daemon_get_device(self, device_number, &dev, error)) {
        if (cdemud_device_unload_disc(CDEMUD_DEVICE(dev), error)) {
            return TRUE;
        }
    }
    
    return FALSE;
}

gboolean cdemud_daemon_device_get_debug_mask (CDEMUD_Daemon *self, gint device_number, gchar *type, gint *dbg_mask, GError **error) {
    SANITY_CHECK();
    /*CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);*/
    GObject *dev = NULL;
        
    if (__cdemud_daemon_get_device(self, device_number, &dev, error)) {
        if (cdemud_device_get_debug_mask(CDEMUD_DEVICE(dev), type, dbg_mask, error)) {
            return TRUE;
        }
    }   
    
    return FALSE;
}

gboolean cdemud_daemon_device_set_debug_mask (CDEMUD_Daemon *self, gint device_number, gchar *type, gint dbg_mask, GError **error) {
    SANITY_CHECK();
    /*CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);*/
    GObject *dev = NULL;
        
    if (__cdemud_daemon_get_device(self, device_number, &dev, error)) {
        if (cdemud_device_set_debug_mask(CDEMUD_DEVICE(dev), type, dbg_mask, error)) {
            return TRUE;
        }
    }
    
    return FALSE;
}


/******************************************************************************\
 *                                 Object init                                *
\******************************************************************************/
/* Our parent class */
static MIRAGE_ObjectClass *parent_class = NULL;

static void __cdemud_daemon_instance_init (GTypeInstance *instance, gpointer g_class) {
    CDEMUD_Daemon *self = CDEMUD_DAEMON(instance);
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    
    /* Set version string */
    _priv->version = g_strdup(PACKAGE_VERSION);
    
    return;
}

static void __cdemud_daemon_dispose (GObject *obj) {
    CDEMUD_Daemon *self = CDEMUD_DAEMON(obj);
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    
    if (_priv->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }
    
    /* Unref all members we own references to */
    g_main_loop_unref(_priv->main_loop);
    _priv->initialized = TRUE; /* Needed for __cdemud_daemon_destroy_devices */
    __cdemud_daemon_destroy_devices(self, NULL);
    g_object_unref(_priv->mirage);
    
    /* Make sure dispose does not run twice. */
    _priv->dispose_has_run = TRUE;
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->dispose(obj);
}

static void __cdemud_daemon_finalize (GObject *obj) {
    CDEMUD_Daemon *self = CDEMUD_DAEMON(obj);
    CDEMUD_DaemonPrivate *_priv = CDEMUD_DAEMON_GET_PRIVATE(self);
    
    /* Close the driver */
    sysfs_close_driver(_priv->sysfs_driver);
    
    /* Free version string */
    g_free(_priv->version);
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void __cdemud_daemon_class_init (gpointer g_class, gpointer g_class_data) {
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    CDEMUD_DaemonClass *klass = CDEMUD_DAEMON_CLASS(g_class);
    
    /* Set parent class */
    parent_class = g_type_class_peek_parent(klass);
    
    /* Register private structure */
    g_type_class_add_private(klass, sizeof(CDEMUD_DaemonPrivate));
    
    /* Initialize GObject methods */
    gobject_class->dispose = __cdemud_daemon_dispose;
    gobject_class->finalize = __cdemud_daemon_finalize;
    
    /* Signal handlers */        
    klass->signals[0] = g_signal_new("daemon-change", G_OBJECT_CLASS_TYPE(klass), (G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED), 0, NULL, NULL, g_cclosure_user_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT, NULL);
    klass->signals[1] = g_signal_new("device-change", G_OBJECT_CLASS_TYPE(klass), (G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED), 0, NULL, NULL, g_cclosure_user_marshal_VOID__INT_INT, G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT, NULL);
    
    return;
}

GType cdemud_daemon_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(CDEMUD_DaemonClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __cdemud_daemon_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(CDEMUD_Daemon),
            0,      /* n_preallocs */
            __cdemud_daemon_instance_init    /* instance_init */
        };
        
        type = g_type_register_static(MIRAGE_TYPE_OBJECT, "CDEMUD_Daemon", &info, 0);
    }
    
    return type;
}
